#include <stdio.h>
#include <syslog.h>
#include <string.h>
#include <signal.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>

char *output = NULL;
int memory_limit = 0;
char *buffer = NULL;
char really_quiet = 0;

#define min(x, y)	( (x) < (y) ? (x) : (y) )

void usage(void)
{
	printf("-o x	file to write to\n");
	printf("	WARNING: use a complete path!\n");
	printf("-m x	size of buffer in bytes (0 = unlimited)\n");
	printf("-z      write immediately to disk when the input-pipe is closed\n");
	printf("	in that case, memtee is just a temporary buffer to avoid\n");
	printf("	(frequent) disk writes\n");
	printf("-q	be quiet\n");
	printf("-Q	be very quiet: do not even output what is read\n");
	printf("-h	this help\n");
	printf("-V	show version and exit\n");
}

void show_version(void)
{
	printf("memtee v" VERSION ", (C) 2005 by folkert@vanheusden.com\n");
}

void sighandler(int sig)
{
	if (buffer)
	{
		FILE *fh = fopen(output, "w");
		if (fh)
		{
			fprintf(fh, "%s", buffer);
			fclose(fh);

			if (!really_quiet)
				syslog(LOG_INFO, "Buffer written to file %s", output);
		}
		else
		{
			syslog(LOG_ERR, "Error creating file %s: %m", output);
		}
	}
	else
	{
		syslog(LOG_WARNING, "Buffer is still empty at this point.");
	}

	if (sig == SIGTERM)
	{
		exit(0);
	}

	signal(sig, sighandler);
}

int main(int argc, char *argv[])
{
	int c;
	char quiet = 0;
	int buffer_size = 0;
	char *buffer_in = NULL;
	int buffer_in_size;
	char exit_at_once = 0;

	while((c = getopt(argc, argv, "o:m:qzQhV")) != -1)
	{
		switch(c) {
		case 'o':
			output = optarg;
			break;

		case 'm':
			memory_limit = atoi(optarg);
			if (memory_limit < 0)
			{
				fprintf(stderr, "Memory limit cannot be less then 0 (%d)\n", memory_limit);
				return 1;
			}
			break;
		case 'z':
			exit_at_once = 1;
			break;
		case 'q':
			quiet = 1;
			break;
		case 'Q':
			quiet = really_quiet = 1;
			break;
		case 'h':
			usage();
			return 1;
		case 'V':
			show_version();
			return 1;
		default:
			fprintf(stderr, "-%c is not understood\n", c);
			return 1;
		}
	}

	if (!output)
	{
		fprintf(stderr, "No outputfile given!\n");
		return 1;
	}

	if (!quiet)
	{
		show_version();
	}

	if (memory_limit > 0)
	{
		buffer = (char *)malloc(memory_limit);
		if (!buffer)
		{
			fprintf(stderr, "Error while allocating %d bytes of memory\n", memory_limit);
			return 1;
		}
	}

	signal(SIGHUP, sighandler);

	buffer_in_size = memory_limit == 0?4096:min(4096, memory_limit);
	buffer_in = (char *)malloc(buffer_in_size + 1);
	if (!buffer_in)
	{
		fprintf(stderr, "Failed to allocate %d bytes of memory\n", buffer_in_size + 1);
		return 1;
	}

	/* read until EOF */
	for(;;)
	{
		char buffer_in[buffer_in_size + 1];
		int rc = read(0, buffer_in, buffer_in_size);

		if (rc == -1)
		{
			if (errno == EINTR || errno == EAGAIN)
				continue;

			fprintf(stderr, "Error reading from pipe: %d/%s\n", errno, strerror(errno));
			break;
		}
		else if (rc == 0)
		{
			if (!quiet)
				printf("EOF reached\n");
			break;
		}

		buffer_in[rc] = 0x00;

		if (!really_quiet)
			printf("%s", buffer_in);

		if (memory_limit)
		{
			int new_size = buffer_size + rc + 1;

			if (new_size > memory_limit)
			{
				int n_to_remove = new_size - memory_limit;

				if ((buffer_size - n_to_remove) > 0)
					memmove(&buffer[0], &buffer[n_to_remove], buffer_size - n_to_remove);
				buffer_size -= n_to_remove;
			}
		}
		else
		{
			buffer = (char *)realloc(buffer, buffer_size + rc + 1);
			if (!buffer)
			{
				fprintf(stderr, "Error growing buffer!\n");
				break;
			}
		}

		memcpy(&buffer[buffer_size], buffer_in, rc);
		buffer_size += rc;
		buffer[buffer_size] = 0x00;
	}

	if (exit_at_once)
	{
		sighandler(SIGTERM);
	}
	else
	{
		signal(SIGTERM, sighandler);

		/* become daemon */
		if (daemon(0, 1) == -1)
		{
			fprintf(stderr, "Error becoming daemon process: %d/%s\n", errno, strerror(errno));
			return 1;
		}

		/* wait for the signal that lets us flush the buffers and exit */
		for(;;)
		{
			pause();
		}
	}

	return 0;
}
